#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import functools
import logging
import pathlib
import typing

import click
from click.decorators import FC

from ._default_values import VALID_LOG_LEVELS
from ._logging import LoggingConfiguration, LoggingState
from ._version import __version__

log = logging.getLogger(__name__)

LOG_STATE_KEY = "logging_state"


def logging_parameters(
    default_configuration: typing.Optional[LoggingConfiguration] = None,
) -> typing.Union[typing.Callable[..., typing.Any], click.Command]:
    """Define a set of logging configuration options to a ``click`` command.

    Args:
        default_configuration: User defined logging default configuration.

    Returns:
        The decorator function object.
    """
    resolved_configuration: LoggingConfiguration
    if not default_configuration:
        resolved_configuration = LoggingConfiguration()
    else:
        resolved_configuration = typing.cast(
            LoggingConfiguration, default_configuration
        )

    def decorator(f: FC) -> FC:
        @click.option(
            "--log-console-enable/--log-console-disable",
            "enable_console_log",
            default=resolved_configuration.enable_console_logging,
            help="Enable or disable console logging.",
            is_flag=True,
            show_default=True,
        )
        @click.option(
            "--log-console-json-enable/--log-console-json-disable",
            "enable_console_json",
            default=resolved_configuration.console_logging.json_enabled,
            help="Enable or disable console JSON logging.",
            is_flag=True,
            show_default=True,
        )
        @click.option(
            "--log-file-enable/--log-file-disable",
            "enable_file_log",
            default=resolved_configuration.enable_file_logging,
            help="Enable or disable file logging.",
            is_flag=True,
            show_default=True,
        )
        @click.option(
            "--log-file-json-enable/--log-file-json-disable",
            "enable_file_json",
            default=resolved_configuration.file_logging.json_enabled,
            help="Enable or disable file JSON logging.",
            is_flag=True,
            show_default=True,
        )
        @click.option(
            "--log-file",
            "log_file",
            default=resolved_configuration.file_logging.log_file_path,
            help="The log file to write to.",
            is_eager=True,
            show_default=True,
            type=click.Path(
                dir_okay=False,
                exists=False,
                file_okay=True,
                path_type=pathlib.Path,
                writable=True,
                readable=True,
            ),
        )
        @click.option(
            "--log-level",
            "log_level",
            default=resolved_configuration.log_level,
            help="Select logging level to apply to all enabled log sinks.",
            show_default=True,
            type=click.Choice(VALID_LOG_LEVELS, case_sensitive=False),
        )
        def wrapper(
            *args: typing.Any,
            enable_console_log: bool,
            enable_console_json: bool,
            enable_file_log: bool,
            enable_file_json: bool,
            log_file: pathlib.Path,
            log_level: str,
            **kwargs: typing.Any,
        ) -> typing.Any:
            ctx = click.get_current_context()
            this_object = ctx.ensure_object(dict)
            if LOG_STATE_KEY not in this_object:
                this_configuration = copy.deepcopy(resolved_configuration)
                this_configuration.enable_console_logging = enable_console_log
                this_configuration.console_logging.json_enabled = (
                    enable_console_json
                )
                this_configuration.enable_file_logging = enable_file_log
                this_configuration.file_logging.json_enabled = enable_file_json
                this_configuration.file_logging.log_file_path = log_file
                this_configuration.log_level = log_level

                this_object[LOG_STATE_KEY] = LoggingState(this_configuration)
                log.info(f"Click logging config version, {__version__}")
            elif not isinstance(this_object, dict):
                raise RuntimeError(
                    "Unable to define logging state since click context.obj is "
                    "not a dictionary"
                )

            return ctx.invoke(
                f,
                *args,
                **kwargs,
            )

        return functools.update_wrapper(typing.cast(FC, wrapper), f)

    return decorator



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_parameters.
#
#  You should have received a copy of the MIT License along with click_logging_parameters.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import logging
import typing

import click
import pytest
from click.testing import CliRunner, Result

from tests.ci.support.click_runner import click_runner  # noqa: F401


def _invoke_command(
    runner, mm, arguments=[]
) -> typing.Tuple[Result, click.Context]:
    click_context: typing.Optional[click.Context] = None

    def _acquire_context(c: click.Context) -> None:
        nonlocal click_context
        click_context = c

    this_main = mm(_acquire_context)
    result = runner.invoke(this_main, arguments)

    return result, click_context


def test_default_logging(click_runner, mock_file_handler, mock_main):
    result, click_context = _invoke_command(click_runner, mock_main)

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state
    assert logging_state.configuration.enable_file_logging
    assert not logging_state.configuration.enable_console_logging
    assert logging_state.configuration.log_level == "warning"


def test_enable_console_logging(click_runner, mock_file_handler, mock_main):
    result, click_context = _invoke_command(
        click_runner, mock_main, ["--log-console-enable"]
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state.configuration.enable_file_logging
    assert logging_state.configuration.enable_console_logging
    assert logging_state.configuration.log_level == "warning"


def test_enable_console_json_logging(
    click_runner, mock_file_handler, mock_main
):
    result, click_context = _invoke_command(
        click_runner,
        mock_main,
        ["--log-console-enable", "--log-console-json-enable"],
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state.configuration.enable_console_logging
    assert logging_state.configuration.console_logging.json_enabled


def test_disable_console_json_logging(
    click_runner, mock_file_handler, mock_main
):
    result, click_context = _invoke_command(
        click_runner,
        mock_main,
        ["--log-console-enable", "--log-console-json-disable"],
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state.configuration.enable_console_logging
    assert not logging_state.configuration.console_logging.json_enabled


def test_disable_file_logging(click_runner, mock_file_handler, mock_main):
    result, click_context = _invoke_command(
        click_runner, mock_main, ["--log-file-disable"]
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert not logging_state.configuration.enable_file_logging
    assert not logging_state.configuration.enable_console_logging
    assert logging_state.configuration.log_level == "warning"


def test_enable_file_json_logging(click_runner, mock_file_handler, mock_main):
    result, click_context = _invoke_command(
        click_runner, mock_main, ["--log-file-enable", "--log-file-json-enable"]
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state.configuration.enable_file_logging
    assert logging_state.configuration.file_logging.json_enabled


def test_disable_file_json_logging(click_runner, mock_file_handler, mock_main):
    result, click_context = _invoke_command(
        click_runner,
        mock_main,
        ["--log-file-enable", "--log-file-json-disable"],
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state.configuration.enable_file_logging
    assert not logging_state.configuration.file_logging.json_enabled


def test_enable_file_logging(click_runner, mock_file_handler, mock_main):
    result, click_context = _invoke_command(
        click_runner, mock_main, ["--log-file-enable"]
    )

    assert result.exit_code == 0

    logging_state = click_context.obj["logging_state"]
    assert logging_state.configuration.enable_file_logging
    assert not logging_state.configuration.enable_console_logging
    assert logging_state.configuration.log_level == "warning"


class TestLogLevel:
    def _do_test(self, this_level: str, mm):
        click_runner = CliRunner()
        result, click_context = _invoke_command(
            click_runner, mm, ["--log-level", this_level]
        )

        if result.exit_code != 0:
            pytest.fail("Valid log level failed test, {0}".format(this_level))

        logging_state = click_context.obj["logging_state"]
        assert logging_state.configuration.log_level == this_level

    def test_critical(self, mock_file_handler, mock_main):
        self._do_test("critical", mock_main)

    def test_error(self, mock_file_handler, mock_main):
        self._do_test("error", mock_main)

    def test_warning(self, mock_file_handler, mock_main):
        self._do_test("warning", mock_main)

    def test_info(self, mock_file_handler, mock_main):
        self._do_test("info", mock_main)

    def test_debug(self, mock_file_handler, mock_main):
        self._do_test("debug", mock_main)

    def test_notset(self, mock_file_handler, mock_main):
        self._do_test("notset", mock_main)

    def test_bad_level(self, click_runner, mock_file_handler, mock_main):
        result, click_context = _invoke_command(
            click_runner, mock_main, ["--log-level", "bad_level"]
        )
        assert result.exit_code != 0

    def test_mixed_case(self, click_runner, mock_file_handler, mock_main):
        result, click_context = _invoke_command(
            click_runner, mock_main, ["--log-level", "Debug"]
        )

        assert result.exit_code == 0

        logging_state = click_context.obj["logging_state"]
        assert logging_state.configuration.log_level == "debug"

    def test_upper_case(self, click_runner, mock_file_handler, mock_main):
        result, click_context = _invoke_command(
            click_runner, mock_main, ["--log-level", "DEBUG"]
        )

        assert result.exit_code == 0

        logging_state = click_context.obj["logging_state"]
        assert logging_state.configuration.log_level == "debug"



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with click_logging_config.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import json
import logging
import pathlib
import re
import typing
from json import JSONDecodeError

import pytest

from click_logging_config._default_values import (
    DEFAULT_LOG_FILE,
    VALID_LOG_LEVELS,
)
from click_logging_config._logging import LoggingConfiguration, LoggingState
from tests.ci.support.directory import change_directory  # noqa: F401

EXPECTED_NONJSON_FORMAT = r"[0-9\-]+T[0-9:.+]+::INFO::.+::{0}"


@pytest.fixture()
def mock_config() -> typing.Callable[
    [typing.Optional[dict]], LoggingConfiguration
]:
    def _apply(
        attributes: typing.Optional[dict] = None,
    ) -> LoggingConfiguration:
        if attributes:
            f = attributes
        else:
            f = {"log_level": "debug"}
        v = LoggingConfiguration.parse_obj(f)

        return v

    return _apply


def test_clean(capsys, caplog, mock_config):
    this_config = mock_config()
    expected_log_level_value = getattr(logging, this_config.log_level.upper())

    root_logger = logging.getLogger()

    assert root_logger.getEffectiveLevel() != expected_log_level_value

    with change_directory() as c:
        under_test = LoggingState(this_config)

    assert root_logger.getEffectiveLevel() == expected_log_level_value


def test_file_logging(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = True

    with change_directory() as c:
        under_test = LoggingState(this_config)

        assert under_test._rotation_handler is not None
        handlers = logging.getLogger().handlers
        assert under_test._rotation_handler in handlers


def test_disabled_file_logging(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False

    with change_directory() as c:
        under_test = LoggingState(this_config)

        assert under_test._rotation_handler is None


def test_file_json_enabled(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = True
    this_config.file_logging.json_enabled = True

    expected_message = f"test message, {__name__}"

    with change_directory() as c:
        file_log = this_config.file_logging.log_file_path
        under_test = LoggingState(this_config)
        log = logging.getLogger(__name__)
        log.info(expected_message)
        log.debug("another message")

        with file_log.open(mode="r") as f:
            content = f.read()

    s = content.splitlines(keepends=False)
    data = [json.loads(x) for x in s]

    assert any(expected_message in x["message"] for x in data)


def test_file_json_disabled(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = True
    this_config.file_logging.json_enabled = False

    expected_message = f"test message, {__name__}"

    with change_directory() as c:
        file_log = this_config.file_logging.log_file_path
        under_test = LoggingState(this_config)
        log = logging.getLogger(__name__)
        log.info(expected_message)

        with file_log.open(mode="r") as f:
            content = f.read()

    try:
        json.loads(content)

        pytest.fail("json file logging is supposed to have been disabled")
    except JSONDecodeError:
        # error reading json is okay because the content should not be json
        pass

    assert any(
        [
            re.match(EXPECTED_NONJSON_FORMAT.format("test message,.*"), x)
            for x in content.splitlines(keepends=False)
        ]
    )


def test_disabled_file_post_construction(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = True

    with change_directory() as c:
        under_test = LoggingState(this_config)

        assert under_test._rotation_handler is not None
        under_test.configuration.enable_file_logging = False
        under_test.set_logging_state()

        assert under_test._rotation_handler is None


def _count_stream_handlers() -> int:
    stream_handlers = [
        isinstance(h, logging.StreamHandler)
        for h in logging.getLogger().handlers
    ]
    return len(stream_handlers)


def test_enabled_console_logging(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False
    this_config.enable_console_logging = True

    sh_before = _count_stream_handlers()
    with change_directory() as c:
        under_test = LoggingState(this_config)

        sh_after = _count_stream_handlers()
        assert sh_after == sh_before + 1


def test_disabled_console_logging(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False
    this_config.enable_console_logging = False
    this_config.console_logging.json_enabled = True

    sh_before = _count_stream_handlers()
    with change_directory() as c:
        under_test = LoggingState(this_config)

        sh_after = _count_stream_handlers()
        # expect no change to stream handlers
        assert sh_after == sh_before


def test_enabled_console_json(capsys, mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False
    this_config.enable_console_logging = True
    this_config.console_logging.json_enabled = True

    expected_message = f"test message, {__name__}"

    with change_directory() as c:
        under_test = LoggingState(this_config)
        log = logging.getLogger(__name__)
        log.info(expected_message)

    content = capsys.readouterr().err

    s = content.splitlines(keepends=False)
    data = [json.loads(x) for x in s]

    assert any(expected_message in x["message"] for x in data)


def test_disabled_console_json(capsys, mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False
    this_config.enable_console_logging = True
    this_config.console_logging.json_enabled = False

    expected_message = f"test message, {__name__}"

    with change_directory() as c:
        under_test = LoggingState(this_config)
        log = logging.getLogger(__name__)
        log.info(expected_message)

    console_content = capsys.readouterr()
    content = console_content.err

    try:
        json.loads(content)

        pytest.fail("json console logging is supposed to have been disabled")
    except JSONDecodeError:
        # error reading json is okay because the content should not be json
        pass

    assert any(
        [
            re.match(EXPECTED_NONJSON_FORMAT.format("test message,.*"), x)
            is not None
            for x in content.splitlines(keepends=False)
        ]
    )


def test_disabled_console_post_construction(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False
    this_config.enable_console_logging = True

    with change_directory() as c:
        under_test = LoggingState(this_config)

        sh_before = _count_stream_handlers()

        under_test.configuration.enable_console_logging = False
        under_test.set_logging_state()

        sh_after = _count_stream_handlers()
        # the exact reduction seems to be unpredictable...
        assert sh_after < sh_before


def test_enabled_console_post_construction(mock_config):
    this_config = mock_config()
    this_config.enable_file_logging = False
    this_config.enable_console_logging = False

    with change_directory() as c:
        under_test = LoggingState(this_config)

        sh_before = _count_stream_handlers()

        under_test.configuration.enable_console_logging = True
        under_test.set_logging_state()

        sh_after = _count_stream_handlers()
        assert sh_after == sh_before + 1


class TestLoggingStateGeneratedLogs:
    @pytest.mark.skip("not sure if this test is valid")
    def test_disable_console_logging(self, capsys, mock_config):
        attributes = {
            "enable_file_logging": False,
            "enable_console_logging": False,
            "log_level": "warning",
        }
        this_config = mock_config(attributes)
        this_config.log_level = "warning"

        captured_err, captured_out, _ = self._do_test(this_config, capsys)

        assert not captured_err
        assert not captured_out

    def _do_test(
        self, this_config: LoggingConfiguration, this_capsys
    ) -> typing.Tuple[str, str, typing.Optional[str]]:
        with change_directory() as temp_dir:
            expected_log_file: pathlib.Path = (
                this_config.file_logging.log_file_path
            )
            under_test = LoggingState(this_config)

            log = logging.getLogger("MockLogging")
            log.debug("debug message")
            log.info("info message")
            log.warning("warning message")
            log.error("error message")
            log.critical("critical message")

            assert (
                under_test.configuration.enable_file_logging
                is this_config.enable_file_logging
            )
            assert (
                under_test.configuration.enable_console_logging
                is this_config.enable_console_logging
            )
            assert (
                under_test.configuration.log_level.upper()
                == this_config.log_level.upper()
            )

            if not this_config.enable_file_logging:
                assert not expected_log_file.exists()

                file_contents = None
            else:
                assert (
                    under_test.configuration.file_logging.log_file_path
                    == DEFAULT_LOG_FILE
                )
                assert expected_log_file.is_file()

                with expected_log_file.open("r") as f:
                    file_contents = f.read()
            (captured_out, captured_err) = this_capsys.readouterr()

        return captured_err, captured_out, file_contents

    def test_enable_console_logging(self, capsys, mock_config):
        attributes = {
            "enable_file_logging": False,
            "enable_console_logging": True,
            "log_level": "warning",
        }
        this_config = mock_config(attributes)

        captured_err, captured_out, _ = self._do_test(this_config, capsys)

        assert not captured_out
        assert "debug" not in captured_err
        assert "info" not in captured_err
        assert "warning message" in captured_err
        assert "error message" in captured_err
        assert "critical message" in captured_err

    def test_disable_file_logging(self, capsys, mock_config):
        attributes = {
            "enable_file_logging": False,
            "enable_console_logging": False,
            "file_logging": {"json_enabled": False},
            "log_level": "warning",
        }
        this_config = mock_config(attributes)

        captured_err, captured_out, file_contents = self._do_test(
            this_config, capsys
        )

        assert not file_contents

    def test_enable_file_logging(self, capsys, mock_config):
        attributes = {
            "enable_file_logging": True,
            "enable_console_logging": False,
            "file_logging": {"json_enabled": False},
            "log_level": "warning",
        }
        this_config = mock_config(attributes)

        captured_err, captured_out, file_contents = self._do_test(
            this_config, capsys
        )

        assert not captured_out

        assert "debug" not in file_contents
        assert "info" not in file_contents
        assert "warning message" in file_contents
        assert "error message" in file_contents
        assert "critical message" in file_contents

    def test_enable_both_file_console(self, capsys, mock_config):
        attributes = {
            "enable_file_logging": True,
            "enable_console_logging": True,
            "file_logging": {"json_enabled": False},
            "log_level": "warning",
        }
        this_config = mock_config(attributes)

        captured_err, captured_out, file_contents = self._do_test(
            this_config, capsys
        )

        assert "debug" not in captured_err
        assert "info" not in captured_err
        assert "warning message" in captured_err
        assert "error message" in captured_err
        assert "critical message" in captured_err

        assert "debug" not in file_contents
        assert "info" not in file_contents
        assert "warning message" in file_contents
        assert "error message" in file_contents
        assert "critical message" in file_contents

    def test_log_levels(self, capsys, mock_config):
        for this_level in VALID_LOG_LEVELS:
            attributes = {
                "enable_file_logging": True,
                "enable_console_logging": True,
                "log_level": this_level,
            }
            this_config = mock_config(attributes)

            captured_err, captured_out, _ = self._do_test(this_config, capsys)

            if captured_out:
                pytest.fail("Captured stdout should be empty")

            if this_level != "notset":
                expected_level = this_level.upper()
                if (expected_level not in captured_err) and (
                    "{0} message".format(expected_level.lower())
                    not in captured_err
                ):
                    pytest.fail(
                        "Log level {0} not present in captured stderr, "
                        "{1}".format(
                            expected_level,
                            captured_err if captured_err else "<empty>",
                        )
                    )
            else:
                # notset enables logging at all levels
                for log_level in VALID_LOG_LEVELS:
                    if log_level != "notset":
                        assert "{0} message".format(log_level) in captured_err



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with click_logging_config.
#  If not, see <https://opensource.org/licenses/MIT>.
#

from click_logging_config._default_values import (
    DEFAULT_CONSOLE_JSON_ENABLED,
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_JSON_ENABLED,
    DEFAULT_FILE_LOGGING_ENABLED,
    DEFAULT_FILE_ROTATION_BACKUPS,
    DEFAULT_FILE_ROTATION_SIZE_MB,
    DEFAULT_LOG_FILE,
    DEFAULT_LOG_LEVEL,
)
from click_logging_config._logging import LoggingConfiguration


def test_default():
    under_test = LoggingConfiguration()

    assert under_test.enable_file_logging is DEFAULT_FILE_LOGGING_ENABLED
    assert under_test.enable_console_logging is DEFAULT_CONSOLE_LOGGING_ENABLED
    assert under_test.log_level == DEFAULT_LOG_LEVEL

    assert (
        under_test.console_logging.json_enabled is DEFAULT_CONSOLE_JSON_ENABLED
    )

    assert under_test.file_logging.json_enabled is DEFAULT_FILE_JSON_ENABLED
    assert under_test.file_logging.log_file_path == DEFAULT_LOG_FILE
    assert (
        under_test.file_logging.file_rotation_size_megabytes
        == DEFAULT_FILE_ROTATION_SIZE_MB
    )
    assert (
        under_test.file_logging.max_rotation_backup_files
        == DEFAULT_FILE_ROTATION_BACKUPS
    )



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import functools
import typing

DecoratorCallable = typing.TypeVar(
    "DecoratorCallable", bound=typing.Callable[..., typing.Any]
)


def f1(f: DecoratorCallable) -> DecoratorCallable:
    print("this is f1 decorator")

    def decorator(p1: str):
        print("this is f1 decorated")
        v = f(p1)

        return f"{v}, f1 decorated"

    return decorator


@f1
def func1(p1: str) -> str:
    v = f"this is func1, {p1}"
    return v


def test_1():
    r = func1("first")

    assert r == "this is func1, first, f1 decorated"


def f2(p2: str) -> DecoratorCallable:
    print(f"this is f2 decorator, {p2}")

    def decorator(f: DecoratorCallable) -> DecoratorCallable:
        print("this is f2 decorated")

        def wrapper(p1: str) -> str:
            print("this is f2 wrapped")
            v = f(p1)

            return f"{v}, f2 decorated, {p2}"

        return wrapper

    return decorator


@f2("f2 arg")
def func2(p1: str) -> str:
    v = f"this is func2, {p1}"
    return v


def test_2():
    r = func2("second")

    assert r == "this is func2, second, f2 decorated, f2 arg"


def f3(p2: str) -> DecoratorCallable:
    print(f"this is f3 decorator, {p2}")

    def decorator(f: DecoratorCallable) -> DecoratorCallable:
        print("this is f3 decorated")

        def wrapper(p3: str, *args) -> str:
            print("this is f3 wrapped")
            v = f(*args)

            return f"{v}, f3 decorated, {p2}, {p3}"

        return functools.update_wrapper(wrapper, f, assigned=("p3"))

    return decorator


@f3("f3p2 arg")
def func3(p1: str) -> str:
    v = f"this is func3, {p1}"
    return v


def test_3():
    r = func3("p3 arg", "p1 arg")

    assert r == "this is func3, p1 arg, f3 decorated, f3p2 arg, p3 arg"



#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.

"""Quick and easy logging parameters for click commands."""

from ._click import logging_parameters  # noqa: F401
from ._logging import LoggingConfiguration, LoggingState  # noqa: F401
from ._version import __version__  # noqa: F401



#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.

"""Logging configuration and state."""

import logging.handlers
import pathlib
import typing

import json_log_formatter  # type: ignore
import pendulum
import pydantic
import pytz

from ._default_values import (
    DEFAULT_CONSOLE_JSON_ENABLED,
    DEFAULT_CONSOLE_LOGGING_ENABLED,
    DEFAULT_FILE_JSON_ENABLED,
    DEFAULT_FILE_LOGGING_ENABLED,
    DEFAULT_FILE_ROTATION_BACKUPS,
    DEFAULT_FILE_ROTATION_SIZE_MB,
    DEFAULT_LOG_FILE,
    DEFAULT_LOG_FORMAT,
    DEFAULT_LOG_LEVEL,
)


class ConsoleLogging(pydantic.BaseModel):
    """Console log configuration parameters."""

    json_enabled: bool = DEFAULT_CONSOLE_JSON_ENABLED


class FileLogging(pydantic.BaseModel):
    """Log file configuration parameters.

    In this context file logs are *always* rotated - the rotation just might be
    at a relatively large file size. As a best practice, not rotating log files
    is considered not particularly useful.
    """

    json_enabled: bool = DEFAULT_FILE_JSON_ENABLED
    log_file_path: pathlib.Path = DEFAULT_LOG_FILE
    file_rotation_size_megabytes: int = DEFAULT_FILE_ROTATION_SIZE_MB
    max_rotation_backup_files: int = DEFAULT_FILE_ROTATION_BACKUPS


class LoggingConfiguration(pydantic.BaseModel):
    """Logging configuration data."""

    log_level: str = DEFAULT_LOG_LEVEL

    enable_console_logging: bool = DEFAULT_CONSOLE_LOGGING_ENABLED
    console_logging: ConsoleLogging = ConsoleLogging()

    enable_file_logging: bool = DEFAULT_FILE_LOGGING_ENABLED
    file_logging: FileLogging = FileLogging()


U = typing.TypeVar("U", bound="Iso8601Formatter")


class Iso8601Formatter(logging.Formatter):
    """Custom formatter with ISO-8601 timestamps."""

    converter: typing.Callable[
        ..., pendulum.DateTime
    ] = pendulum.from_timestamp  # type: ignore [assignment]

    def formatTime(
        self: U,
        record: logging.LogRecord,
        datefmt: typing.Optional[str] = None,
        timezone: typing.Optional[str] = None,
    ) -> str:
        """Generate formatted time."""
        if timezone:
            v = Iso8601Formatter.converter(
                record.created,
                tz=pytz.timezone(timezone),
            ).isoformat()
        else:
            v = Iso8601Formatter.converter(
                record.created,
            ).isoformat()

        return v


T = typing.TypeVar("T", bound="LoggingState")


class LoggingState:
    """Logging configuration parameters."""

    configuration: LoggingConfiguration
    _console_handler: typing.Optional[logging.StreamHandler]
    _rotation_handler: typing.Optional[logging.handlers.RotatingFileHandler]

    def __init__(
        self: T,
        logging_configuration: LoggingConfiguration,
    ) -> None:
        """Construct ``LoggingState`` object.

        Args:
            logging_configuration: Logging configuration to be applied.
        """
        self.configuration = logging_configuration
        self._console_handler = None
        self._rotation_handler = None

        self.set_logging_state()

    def set_logging_state(self: T) -> None:
        """Apply the logging state from configuration."""
        root_logger = logging.getLogger()
        self.__set_log_level(root_logger)
        self.__set_file_logging(root_logger)
        self.__set_console_logging(root_logger)

    def __level_value(self: T) -> int:
        """Convert log level text to a ``logging`` framework integer."""
        return getattr(logging, self.configuration.log_level.upper())

    def __set_file_logging(self: T, root_logger: logging.Logger) -> None:
        """Enable or disable file logging.

        Args:
            root_logger: Root logger to modify.
        """
        if self.configuration.enable_file_logging:
            # No change if a rotation handler already exists.
            if not self._rotation_handler:
                this_handler = logging.handlers.RotatingFileHandler(
                    str(self.configuration.file_logging.log_file_path),
                    backupCount=(
                        self.configuration.file_logging.max_rotation_backup_files  # noqa: E501
                    ),
                    maxBytes=(  # noqa: E501
                        self.configuration.file_logging.file_rotation_size_megabytes  # noqa: E501
                        * (1024**2)
                    ),
                )
                this_handler.setLevel(self.__level_value())
                if self.configuration.file_logging.json_enabled:
                    this_handler.setFormatter(
                        json_log_formatter.VerboseJSONFormatter()
                    )
                else:
                    this_handler.setFormatter(
                        Iso8601Formatter(fmt=DEFAULT_LOG_FORMAT)
                    )

                self._rotation_handler = this_handler

                root_logger.addHandler(self._rotation_handler)
        elif self._rotation_handler:
            self._rotation_handler.flush()
            self._rotation_handler.close()
            root_logger.removeHandler(self._rotation_handler)
            self._rotation_handler = None
        # else self._rotation_handler is None and not self.enable_file_logging
        # so do nothing

    def __set_console_logging(self: T, root_logger: logging.Logger) -> None:
        """Enable or disable console logging.

        Args:
            root_logger: Root logger to modify.
        """
        # def remove_stream_handlers() -> None:
        #     nonlocal root_logger
        #     for this_handler in root_logger.handlers:
        #         if type(this_handler) == logging.StreamHandler:
        #             root_logger.removeHandler(this_handler)

        # Have to remove ALL stream handlers to fully disable console
        # logging.
        # remove_stream_handlers()
        if self.configuration.enable_console_logging:
            self._console_handler = logging.StreamHandler()
            self._console_handler.setLevel(self.__level_value())
            if self.configuration.console_logging.json_enabled:
                self._console_handler.setFormatter(
                    json_log_formatter.VerboseJSONFormatter()
                )
            else:
                self._console_handler.setFormatter(
                    Iso8601Formatter(fmt=DEFAULT_LOG_FORMAT)
                )

            root_logger.addHandler(self._console_handler)
        elif self._console_handler:
            self._console_handler.flush()
            self._console_handler.close()
            root_logger.removeHandler(self._console_handler)
            self._console_handler = None
        # else self._console_handler is None and not
        # self.enable_console_logging so do nothing

    def __set_log_level(self: T, root_logger: logging.Logger) -> None:
        """Set log level on any existing handlers.

        Args:
            root_logger: Root logger to modify.
        """
        for this_handler in root_logger.handlers:
            this_handler.setLevel(self.__level_value())
        # Ensure that the logging level propagates to any subsequently created
        # handlers by setting the root logger level as well.
        root_logger.setLevel(self.__level_value())



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_parameters.
#
#  You should have received a copy of the MIT License along with click_logging_parameters.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import typing

import click
import pytest

import click_logging_config._logging
from click_logging_config import LoggingConfiguration, logging_parameters


@pytest.fixture()
def mock_file_handler(mocker):
    mocker.patch(
        "click_logging_config._logging.logging.handlers.RotatingFileHandler"
    )
    mocker.patch.object(click_logging_config._logging.logging, "getLogger")


CallbackCallable = typing.Callable[[click.Context], None]


@pytest.fixture()
def mock_main() -> typing.Callable[[CallbackCallable], click.BaseCommand]:
    def _apply(
        callback: CallbackCallable,
        user_defaults: typing.Optional[LoggingConfiguration] = None,
    ) -> click.BaseCommand:
        @click.command()
        @click.option("--p", default=None, type=str)
        @logging_parameters(user_defaults)
        def this_main(p: typing.Optional[str]) -> None:
            c = click.get_current_context()
            c.obj["p"] = p
            callback(c)

        return this_main

    return _apply



#
#  Copyright (c) 2022 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.
#

import copy
import logging
import pathlib
import typing

import click

from click_logging_config._click import LoggingConfiguration, logging_parameters
from click_logging_config._logging import FileLogging, LoggingState
from tests.ci.support.click_runner import click_runner  # noqa: F401


class TestLoggingParameters:
    def test_logging_option(self, capsys, click_runner):
        """Using a logging option."""
        this_context: typing.Optional[click.Context] = None
        user_default_attributes = {
            "enable_console_logging": False,
            "enable_file_logging": False,
            "file_logging": {
                "log_file_path": pathlib.Path("some_file.log"),
                "log_level": logging.DEBUG,
                "file_rotation_size_megabytes": 5,
                "max_rotation_backup_files": 50,
            },
        }
        expected_attributes = {
            "enable_console_logging": True,
            "enable_file_logging": False,
            "file_logging": {
                "log_file_path": pathlib.Path("some_file.log"),
                "log_level": logging.DEBUG,
                "file_rotation_size_megabytes": 5,
                "max_rotation_backup_files": 50,
            },
        }
        expected = LoggingConfiguration.parse_obj(expected_attributes)

        @click.command()
        @click.option(
            "--parm",
            default="a value",
            type=str,
        )
        @logging_parameters(
            LoggingConfiguration.parse_obj(user_default_attributes)
        )
        def mock_main(
            parm: str,
        ):
            nonlocal this_context
            this_context = click.get_current_context()

            assert parm == "a value"

        result = click_runner.invoke(mock_main, ["--log-console-enable"])

        assert result.exit_code == 0

        this_state = this_context.obj["logging_state"]
        assert (
            this_state.configuration.file_logging.log_file_path
            == expected.file_logging.log_file_path
        )
        assert this_state.configuration.log_level == expected.log_level
        assert (
            this_state.configuration.enable_file_logging
            is expected.enable_file_logging
        )
        assert (
            this_state.configuration.enable_console_logging
            is expected.enable_console_logging
        )

    def test_no_user_ctx(self, capsys, click_runner):
        """User does not use ``@click.pass_context decorator``.

        Logging parameters continues to be able to consume click context.
        """
        this_context: typing.Optional[click.Context] = None
        expected = LoggingConfiguration.parse_obj(
            {
                "enable_console_logging": False,
                "enable_file_logging": False,
                "file_logging": FileLogging.parse_obj(
                    {
                        "log_file_path": pathlib.Path("some_file.log"),
                        "log_level": logging.DEBUG,
                        "file_rotation_size_megabytes": 5,
                        "max_rotation_backup_files": 50,
                    },
                ),
            },
        )

        @click.command()
        @click.option(
            "--mock-p",
            default="a value",
            type=str,
        )
        @logging_parameters(expected)
        def mock_main(
            mock_p: str,
        ):
            nonlocal this_context
            this_context = copy.deepcopy(click.get_current_context())

            assert mock_p == "mp"

        result = click_runner.invoke(mock_main, ["--mock-p", "mp"])

        assert result.exit_code == 0

        this_state = this_context.obj["logging_state"]
        assert (
            this_state.configuration.file_logging.log_file_path
            == expected.file_logging.log_file_path
        )
        assert this_state.configuration.log_level == expected.log_level
        assert (
            this_state.configuration.enable_file_logging
            is expected.enable_file_logging
        )
        assert (
            this_state.configuration.enable_console_logging
            is expected.enable_console_logging
        )

    def test_user_ctx(self, capsys, click_runner, mocker):
        """User uses ``@click.pass_context decorator``.

        The user consumption of click context does not interfere with
        logging parameters context.
        """
        this_context: typing.Optional[click.Context] = None
        expected = LoggingConfiguration.parse_obj(
            {
                "enable_console_logging": False,
                "enable_file_logging": False,
                "file_logging": FileLogging.parse_obj(
                    {
                        "log_file_path": pathlib.Path("some_file.log"),
                        "log_level": logging.DEBUG,
                        "file_rotation_size_megabytes": 5,
                        "max_rotation_backup_files": 50,
                    }
                ),
            }
        )

        @click.command()
        @click.pass_context
        @click.option(
            "--mock-p",
            default="a value",
            type=str,
        )
        @logging_parameters(expected)
        def mock_main(
            ctx: click.Context,
            mock_p: str,
        ):
            nonlocal this_context
            this_context = copy.deepcopy(ctx)

            assert mock_p == "mp"

        result = click_runner.invoke(mock_main, ["--mock-p", "mp"])

        assert result.exit_code == 0

        this_state = this_context.obj["logging_state"]
        assert (
            this_state.configuration.file_logging.log_file_path
            == expected.file_logging.log_file_path
        )
        assert this_state.configuration.log_level == expected.log_level
        assert (
            this_state.configuration.enable_file_logging
            is expected.enable_file_logging
        )
        assert (
            this_state.configuration.enable_console_logging
            is expected.enable_console_logging
        )



#  Copyright (c) 2020 Russell Smiley
#
#  This file is part of click_logging_config.
#
#  You should have received a copy of the MIT License along with build_harness.
#  If not, see <https://opensource.org/licenses/MIT>.

"""Default values."""

import pathlib

DEFAULT_RELEASE_ID = "0.0.0"

DEFAULT_CONSOLE_JSON_ENABLED = False
DEFAULT_CONSOLE_LOGGING_ENABLED = False

DEFAULT_FILE_JSON_ENABLED = True
DEFAULT_FILE_LOGGING_ENABLED = True
DEFAULT_FILE_ROTATION_BACKUPS = 10
DEFAULT_FILE_ROTATION_SIZE_MB = 1

DEFAULT_LOG_FILE = pathlib.Path("this.log")
DEFAULT_LOG_FORMAT = "%(asctime)s::%(levelname)s::%(name)s::%(message)s"
DEFAULT_LOG_LEVEL = "warning"

VALID_LOG_LEVELS = ["critical", "error", "warning", "info", "debug", "notset"]
